@@ -1,6 +1,6 @@
 import numpy as np
 import matplotlib.pyplot as plt
-from kawin.EffectiveDiffusion import effectiveDiffusionDistance, noDiffusionDistance
+from kawin.EffectiveDiffusion import EffectiveDiffusionFunctions
 from kawin.ShapeFactors import ShapeFactor
 from kawin.ElasticFactors import StrainEnergy
 from kawin.GrainBoundaries import GBFactors
@@ -51,10 +51,13 @@ class PrecipitateBase:
             print('\tSetting t0 to {:.3e}'.format(self.t0))
         
         #Time variables
-        self._timeIncrementCheck = self._noCheckDT
+        self.adaptiveTimeStepping(True)
 
         #Predefined constraints, these can be set if they make the simulation unstable
         self._defaultConstraints()
+
+        #Stopping conditions
+        self.clearStoppingConditions()
             
         #Composition array
         self.elements = elements
@@ -74,7 +77,8 @@ class PrecipitateBase:
         self.RdrivingForceLimit = np.zeros(len(self.phases), dtype=np.float32)
         self.shapeFactors = [ShapeFactor() for i in self.phases]
         self.theta = 2 * np.ones(len(self.phases), dtype=np.float32)
-        self.effDiffDistance = effectiveDiffusionDistance
+        self.effDiffFuncs = EffectiveDiffusionFunctions()
+        self.effDiffDistance = self.effDiffFuncs.effectiveDiffusionDistance
         self.infinitePrecipitateDiffusion = [True for i in self.phases]
         self.dTemp = 0
         self.iterationSinceTempChange = 0
@@ -129,7 +133,7 @@ class PrecipitateBase:
     def reset(self):
         '''
         Resets simulation results
-        This does not reset the model parameters
+        This does not reset the model parameters, however, it will clear any stopping conditions
         '''
         self._resetArrays()
         self.xComp[0] = self.xInit
@@ -353,23 +357,49 @@ class PrecipitateBase:
         '''
         if adaptive:
             self._timeIncrementCheck = self._checkDT
+            self._postTimeIncrementCheck = self._postCheckDT
         else:
             self._timeIncrementCheck = self._noCheckDT
+            self._postTimeIncrementCheck = self._noPostCheckDT
+
+    def _calculateDT(self, i, fraction):
+        '''
+        Calculates DT as a fraction of the total simulation time
+        '''
+        if self.linearTimeSpacing:
+            dt = fraction*(self.tf - self.t0)
+        else:
+            dt = self.time[i] * (np.exp(fraction*np.log(self.tf / self.t0)) - 1)
+        return dt
 
     def _defaultConstraints(self):
         '''
         Default values for contraints
         '''
-        self.minRadius = 5e-10
+        self.minRadius = 3e-10
         self.maxTempChange = 1
-        self.maxDTFraction = 0.1
+
+        self.maxDTFraction = 1e-2
         self.minDTFraction = 1e-5
+
+        self.checkTemperature = True
+        self.maxNonIsothermalDT = 1
+
+        self.checkPSD = True
         self.maxDissolution = 0.01
+
+        self.checkRcrit = True
         self.maxRcritChange = 0.01
+
+        self.checkNucleation = True
         self.maxNucleationRateChange = 0.5
         self.minNucleationRate = 1e-5
+
+        self.checkVolumePre = True
+        self.checkVolumePost = True
         self.maxVolumeChange = 0.001
-        self.maxNonIsothermalDT = 1
+        
+        self.checkComposition = True
         self.maxCompositionChange = 0.001
 
     def setConstraints(self, **kwargs):
@@ -378,22 +408,35 @@ class PrecipitateBase:
 
         TODO: the following constraints are not implemented
             maxDTFraction
-            minDTFraction
-            maxRcritChange - will need to account for special case that driving force becomes negative
+            maxRcritChange - this is somewhat implemented but disabled by default
 
         Possible constraints:
         ---------------------
-        minRadius - minimum radius to be considered a precipitate - default is 5e-10 m^3
-        maxTempChange - maximum temperature change before lookup table is updated (only for Euler in binary case) - default is 1 K
-        maxDTFraction - maximum time increment allowed as a fraction of total simulation time - default is 0.1
-        minDTFraction - minimum time increment allowed as a fraction of total simulation time - default is 1e-5
-        maxDissolution - maximum volume fraction of precipitates allowed to dissolve in a single time step - default is 0.01
-        maxRcritChange - maximum change in critical radius (as a fraction) per single time step - default is 0.01
-        maxNucleationRateChange - maximum change in nucleation rate (on log scale) per single time step - default is 0.5
-        minNucleationRate - minimum nucleation rate to be considered for checking time intervals - 1e-5
-        maxVolumeChange - maximum absolute value that volume fraction can change per single time step - default is 0.001
-        maxNonIsothermalDT - maximum time step when temperature is changing - default is 1 s
-        maxCompositionChange - maximum change in composition in single time step - default is 0.01
+        minRadius - minimum radius to be considered a precipitate (1e-10 m)
+        maxTempChange - maximum temperature change before lookup table is updated (only for Euler in binary case) (1 K)
+
+        maxDTFraction - maximum time increment allowed as a fraction of total simulation time (0.1)
+        minDTFraction - minimum time increment allowed as a fraction of total simulation time (1e-5)
+
+        checkTemperature - checks max temperature change (True)
+        maxNonIsothermalDT - maximum time step when temperature is changing (1 second)
+
+        checkPSD - checks maximum growth rate for particle size distribution (True)
+        maxDissolution - maximum relative volume fraction of precipitates allowed to dissolve in a single time step (0.01)
+
+        checkRcrit - checks maximum change in critical radius (False)
+        maxRcritChange - maximum change in critical radius (as a fraction) per single time step (0.01)
+
+        checkNucleation - checks maximum change in nucleation rate (True)
+        maxNucleationRateChange - maximum change in nucleation rate (on log scale) per single time step (0.5)
+        minNucleationRate - minimum nucleation rate to be considered for checking time intervals (1e-5)
+
+        checkVolumePre - checks maximum estimated volume change (True)
+        checkVolumePost - checks maximum calculated volume change (True)
+        maxVolumeChange - maximum absolute value that volume fraction can change per single time step (0.001)
+
+        checkComposition - checks maximum change in composition (True)
+        maxCompositionChange - maximum change in composition in single time step (0.01)
         '''
         for key, value in kwargs.items():
             setattr(self, key, value)
@@ -562,7 +605,6 @@ class PrecipitateBase:
         self.VmBeta[index] = Vm
         self.VaBeta[index] = atomsPerCell * self.VmBeta[index] / self.avo
         self.atomsPerCellBeta[index] = atomsPerCell
-        self.Rmin[index] = self.minRadius
         
     def setVaBeta(self, Va, atomsPerCell, phase = None):
         '''
@@ -581,7 +623,6 @@ class PrecipitateBase:
         self.VaBeta[index] = Va
         self.VmBeta[index] = self.VaBeta[index] * self.avo / atomsPerCell
         self.atomsPerCellBeta[index] = atomsPerCell
-        self.Rmin[index] = self.minRadius
         
     def setUnitCellBeta(self, a, atomsPerCell, phase = None):
         '''
@@ -600,7 +641,6 @@ class PrecipitateBase:
         self.VaBeta[index] = a**3
         self.VmBeta[index] = self.VaBeta[index] * self.avo / atomsPerCell
         self.atomsPerCellBeta[index] = atomsPerCell
-        self.Rmin[index] = self.minRadius
 
     def setNucleationDensity(self, grainSize = 100, aspectRatio = 1, dislocationDensity = 5e12, bulkN0 = None):
         '''
@@ -946,9 +986,93 @@ class PrecipitateBase:
             If False, will calculate correction factor from Chen, Jeppson and Agren (2008)
         '''
         if neglect:
-            self.effDiffDistance = noDiffusionDistance
+            self.effDiffDistance = self.effDiffFuncs.noDiffusionDistance
         else:
-            self.effDiffDistance = effectiveDiffusionDistance
+            self.effDiffDistance = self.effDiffFuncs.effectiveDiffusionDistance
+
+    def addStoppingCondition(self, variable, condition, value, phase = None, element = None, mode = 'or'):
+        '''
+        Adds condition to stop simulation when condition is met
+
+        Parameters
+        ----------
+        variable : str
+            Variable to set condition for, options are 
+                'Volume Fraction'
+                'Average Radius'
+                'Driving Force'
+                'Nucleation Rate'
+                'Precipitate Density'
+        condition : str
+            Operator for condition, options are
+                'greater than' or '>'
+                'less than' or '<'
+        value : float
+            Value for condition
+        phase : str (optional)
+            Phase to consider (defaults to first precipitate in list)
+        element : str (optional)
+            For 'Composition', element to consider for condition (defaults to first element in list)
+        mode : str (optional)
+            How the condition will be handled
+                'or' (default) - at least one condition in this mode needs to be met before stopping
+                'and' - all conditions in this mode need to be met before stopping
+                    This will also record the times each condition is met
+
+        Example
+        model.addStoppingCondition('Volume Fraction', '>', 0.002, 'beta')
+            will add a condition to stop simulation when the volume fraction of the 'beta'
+            phase becomes greater than 0.002
+        '''
+        index = self.phaseIndex(phase)
+
+        if self._stoppingConditions is None:
+            self._stoppingConditions = []
+            self.stopConditionTimes = []
+            self._stopConditionMode = []
+
+        standardLabels = {
+            'Volume Fraction': 'betaFrac',
+            'Average Radius': 'avgR',
+            'Driving Force': 'dGs',
+            'Nucleation Rate': 'nucRate',
+            'Precipitate Density': 'precipitateDensity',
+        }
+        otherLabels = ['Composition']
+
+        if variable in standardLabels:
+            if 'greater' in condition or '>' in condition:
+                cond = lambda self, i, p = index, var=standardLabels[variable] : getattr(self, var)[p,i] > value
+            elif 'less' in condition or '<' in condition:
+                cond = lambda self, i, p = index, var=standardLabels[variable] : getattr(self, var)[p,i] < value
+        else:
+            if variable == 'Composition':
+                eIndex = 0 if element is None else self.elements.index(element)
+                if 'greater' in condition or '>' in condition:
+                    if self.numberOfElements > 1:
+                        cond = lambda self, i, e = eIndex, var='xComp' : getattr(self, var)[i, e] > value
+                    else:
+                        cond = lambda self, i, var='xComp' : getattr(self, var)[i] > value
+                elif 'less' in condition or '<' in condition:
+                    if self.numberOfElements > 1:
+                        cond = lambda self, i, e = eIndex, var='xComp' : getattr(self, var)[i, e] < value
+                    else:
+                        cond = lambda self, i, var='xComp' : getattr(self, var)[i] < value
+
+        self._stoppingConditions.append(cond)
+        self.stopConditionTimes.append(-1)
+        if mode == 'and':
+            self._stopConditionMode.append(False)
+        else:
+            self._stopConditionMode.append(True)
+
+    def clearStoppingConditions(self):
+        '''
+        Clears all stopping conditions
+        '''
+        self._stoppingConditions = None
+        self.stopConditionTimes = None
+        self._stopConditionMode = None
 
     def printModelParameters(self):
         '''
@@ -973,6 +1097,8 @@ class PrecipitateBase:
             #Set nucleation density assuming grain size of 100 um and dislocation density of 5e12 m/m3 (Thermocalc default)
             print('Nucleation density not set.\nSetting nucleation density assuming grain size of {:.0f} um and dislocation density of {:.0e} #/m2'.format(100, 5e12))
             self.setNucleationDensity(100, 1, 5e12)
+        for p in range(len(self.phases)):
+            self.Rmin[p] = self.minRadius
         self._getNucleationDensity()
         self._setGBfactors()
         self._setupStrainEnergyFactors()
@@ -1019,9 +1145,36 @@ class PrecipitateBase:
         
         #While loop since number of steps may change with adaptive time stepping
         i = 1
-        while i < self.steps:
+        stopCondition = False
+        while i < self.steps and not stopCondition:
             self._iterate(i)
 
+            #Apply stopping condition
+            if self._stoppingConditions is not None:
+                andConditions = True
+                numberOfAndConditions = 0
+                orConditions = False
+                for s in range(len(self._stoppingConditions)):
+                    #Record time if stopping condition is met
+                    conditionResult = self._stoppingConditions[s](self, i)
+                    if conditionResult and self.stopConditionTimes[s] == -1:
+                        self.stopConditionTimes[s] = self.time[i]
+
+                    #If condition mode is 'or'
+                    if self._stopConditionMode[s]:
+                        orConditions = orConditions or conditionResult
+                    #If condition mode is 'and'
+                    else:
+                        andConditions = andConditions and conditionResult
+                        numberOfAndConditions += 1
+
+                #If there are no 'and' conditions, andConditions will be True
+                #Set to False so andConditions will not stop the model unneccesarily
+                if numberOfAndConditions == 0:
+                    andConditions = False
+
+                stopCondition = andConditions or orConditions
+
             #Print current variables
             if i % vIt == 0 and verbose:
                 self._printOutput(i)
@@ -1057,22 +1210,21 @@ class PrecipitateBase:
         self.dGs[p, i], _ = self.dG[p](self.xComp[i-1], self.T[i])
         self.dGs[p, i] /= self.VmBeta[p]
 
-        #Temporary, may add way to solve for aspect ratio by minimizing free energy between surface and elastic energy
+        #Add strain energy for spherical shape, use previous critical radius
+        #This should still be correct even if the interfacial energy dominates at small radii since the aspect ratio may be near 1
         self.dGs[p, i] -= self.strainEnergy[p].strainEnergy(self.shapeFactors[p].normalRadii(self.Rcrit[p, i-1]))
 
         if self.dGs[p, i] < 0:
             return self._noDrivingForce(p, i)
 
         #Only do this if there is some parent phase left (brute force solution for to avoid numerical errors)
-        #if self.betaFrac[p, i-1] < 1 and self.xComp[i-1] > 0:
         if self.betaFrac[p, i-1] < 1:
             
             #Calculate critical radius
-            #For bulk or dislocation nucleation sites, the precipitate can be any shape,
-            # so we need to solve for the critical radius in case the aspect ratio is not constant
+            #For bulk or dislocation nucleation sites, use previous critical radius to get aspect ratio
             if self.GB[p].nucleationSiteType == GBFactors.BULK or self.GB[p].nucleationSiteType == GBFactors.DISLOCATION:
-                #self.Rcrit[p, i] = self.shapeFactors[p].findRcrit(2 * self.gamma[p] / self.dGs[p, i], 20 * self.gamma[p] / self.dGs[p, i])
                 self.Rcrit[p, i] = 2 * self.shapeFactors[p].thermoFactor(self.Rcrit[p, i-1]) * self.gamma[p] / self.dGs[p, i]
+                #self.Rcrit[p, i] = 2 * self.gamma[p] / self.dGs[p, i]
                 if self.Rcrit[p, i] < self.Rmin[p]:
                     self.Rcrit[p, i] = self.Rmin[p]
                 
@@ -1089,17 +1241,6 @@ class PrecipitateBase:
             #Calculate nucleation rate
             Z = self._Zeldovich(p, i)
             self.betas[p,i] = self._Beta(p, i)
-
-            #I find just calculating tau assuming steady state seems to work better
-            #I think it's because we're integrating the incubation term, where the thermocalc method of finding tau may already account for it
-            #tau = 1 / (self.theta[p] * (np.pi * self.betas[p,i] * Z**2))
-                
-            #If tau becomes super large, then add delay to the incubation time
-            #if tau > 10 * self.time[-1] and self.incubationOffset[p] == 0:
-            #If driving force disappears, then set the offset time for incubation
-            #if self.dGs[p,i] < 0:
-            #    self.incubationOffset[p] = self.time[i-1]
-            #    self.incubationOffset[p] = np.amax([i-1, 0])
                 
             #Incubation time, either isothermal or nonisothermal
             self.incubationSum[p] = self._incubation(Z, p, i)
@@ -1113,7 +1254,13 @@ class PrecipitateBase:
             
     def _noCheckDT(self, i):
         '''
-        Default time increment function if not implemented
+        Default for no adaptive time stepping
+        '''
+        pass
+
+    def _noPostCheckDT(self, i):
+        '''
+        Default for no adaptive time stepping
         '''
         pass
 
@@ -1123,6 +1270,12 @@ class PrecipitateBase:
         '''
         pass
 
+    def _postCheckDT(self, i):
+        '''
+        Default time increment function if implement (which is no implementation)
+        '''
+        pass
+
     def _noDrivingForce(self, p, i):
         '''
         Set everything to 0 if there is no driving force for precipitation
@@ -1194,8 +1347,17 @@ class PrecipitateBase:
             tau = self.time[int(self.incubationOffset[p]):-1][signChange][0] - self.time[int(self.incubationOffset[p])]
 
         return np.exp(-tau / (self.time[i] - self.time[int(self.incubationOffset[p])]))
+
+    def _setNucleateRadius(self, i):
+        for p in range(len(self.phases)):
+            #If nucleates form, then calculate radius of precipitate
+            #Radius is set slightly larger so preciptate 
+            if self.nucRate[p,i]*(self.time[i]-self.time[i-1]) >= 1 and self.Rcrit[p, i] >= self.Rmin[p]:
+                self.Rad[p, i] = self.Rcrit[p, i] + 0.5 * np.sqrt(self.kB * self.T[i] / (np.pi * self.gamma[p]))
+            else:
+                self.Rad[p, i] = 0
     
-    def plot(self, axes, variable, bounds = None, timeUnits = 's', *args, **kwargs):
+    def plot(self, axes, variable, bounds = None, timeUnits = 's', radius='spherical', *args, **kwargs):
         '''
         Plots model outputs
         
@@ -1217,6 +1379,14 @@ class PrecipitateBase:
                     
         bounds : tuple (optional)
             Limits on the x-axis (float, float) or None (default, this will set bounds to (initial time, final time))
+        timeUnits : str (optional)
+            Plot time dependent variables per seconds ('s'), minutes ('m') or hours ('h')
+        radius : str (optional)
+            For non-spherical precipitates, plot the Average Radius by the -
+                Equivalent spherical radius ('spherical')
+                Short axis ('short')
+                Long axis ('long')
+            Note: Total Average Radius and Volume Average Radius will still use the equivalent spherical radius
         *args, **kwargs - extra arguments for plotting
         '''
         timeScale = 1
@@ -1370,8 +1540,17 @@ class PrecipitateBase:
                 plotVariable = self.Rcrit
             elif variable == 'Average Radius':
                 plotVariable = self.avgR
+                for p in range(len(self.phases)):
+                    if self.GB[p].nucleationSiteType == self.GB[p].BULK or self.GB[p].nucleationSiteType == self.GB[p].DISLOCATION:
+                        if radius != 'spherical':
+                            plotVariable[p] /= self.shapeFactors[p].eqRadiusFactor(self.avgR[p])
+                        if radius == 'long':
+                            plotVariable[p] *= self.avgAR[p]
+                    else:
+                        plotVariable[p] *= self._GBareaRemoval(p)
+
             elif variable == 'Volume Average Radius':
-                plotVariable = np.cbrt(self.betaFrac / self.precipitateDensity)
+                plotVariable = np.cbrt(self.betaFrac / self.precipitateDensity / (4/3*np.pi))
             elif variable == 'Aspect Ratio':
                 plotVariable = self.avgAR
             elif variable == 'Driving Force':
