@@ -8,24 +8,56 @@ class PopulationBalanceModel:
     Class for handling particle size distributions (PSD)
         This include time evolution, moments and graphing
 
+    NOTE: more formally, the PSD is defined such that the number of particles of size R in a volume
+        is the integral of the PSD from R-1/2 to R+1/2
+        For the discrete PSD, n_i = PSD_i * (R+1/2 - R-1/2)
+        Here, we just store the PSD as the number of particles (n_i), but the formulation for 
+        growth rates and moments are the same since the factor (R+1/2 - R-1/2) cancels out
+
     Parameters
     ----------
     cMin : float (optional)
-        Lower bound of PSD (defaults at 0)
+        Lower bound of PSD (defaults at 1e-10)
     cMax : float (optional)
-        Upper bound of PSD (defaults at 100)
+        Upper bound of PSD (defaults at 1e-9)
     bins : int (optional)
-        Number of bins (defaults at 100)
-    linearSpacing : bool (optional)
-        If True (default), will store PSD on a linear scale
-        Else, will store PSD on a logarithmic scale
+        Initial number of bins (defaults at 150)
+    minBins : int (optional)
+        Minimum number of bins over an order of magnitude (defaults at 100)
+    maxBins : int (optional)
+        Maximum number of bins over an order of magnitude (defaults at 200)
+
+    Attributes
+    ----------
+    originalMin : float
+        Minimum bin size initially set
+    min : float
+        Current minimum bin size (this will almost always be equal to originalMin)
+    originalMax : float
+        Maximum bin size initially set
+    max : float
+        Current maximum bin size
+    bins : int
+        Default number of bins
+    minBins : int
+        Minimum number of allowed bins
+    maxBins : int
+        Maximum number of allowed bins
+    PSD : array
+        Particle size distribution
+    PSDsize : array
+        Average radius of each PSD size class
+    PSDbounds : array
+        Radius at the bounds of each size class - length of array is len(PSD)+1
     '''
-    def __init__(self, cMin = 0, cMax = 100, bins = 100, linearSpacing = True):
+    def __init__(self, cMin = 1e-10, cMax = 1e-9, bins = 150, minBins = 100, maxBins = 200):
+        self.originalMin = cMin
+        self.originalMax = np.amax([10*self.originalMin, cMax])
+        self.min = self.originalMin
+        self.max = self.originalMax
+
         self.originalBins = bins
-        self.bins = bins
-        self.min = cMin
-        self.max = cMax
-        self.linear = linearSpacing
+        self.setBinConstraints(bins, minBins, maxBins)
         
         self.reset()
 
@@ -37,76 +69,32 @@ class PopulationBalanceModel:
         #Hidden variable for use in KWNEuler when adaptive time stepping is enabled
         #This allows for PSD to revert to its previous value if a time constraint is not met
         self._prevPSD = np.zeros(self.bins)
-        
-    def reset(self):
-        '''
-        Resets the PSD to 0
-        This will remove any size classes that were added since initialization
-        '''
-        self.bins = self.originalBins
-        if self.linear:
-            self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
-            self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
-        else:
-            self.PSDbounds = np.logspace(self.min, self.max, self.bins+1)
-            self.PSDsize = np.power(10, 0.5 * (np.log(self.PSDbounds[1:]) + np.log(self.PSDbounds[:-1])))
-            
-        self.PSD = np.zeros(self.bins)
+        self._prevPSDbounds = np.zeros(self.bins)
 
-    def revert(self):
-        self.PSD = copy.copy(self._prevPSD)
+        self._adaptiveBinSize = True
 
-    def changeSizeClasses(self, cMin, cMax, bins, linearSpacing = True, resetPSD = True):
+    def setAdaptiveBinSize(self, adaptive):
         '''
-        Changes the size classes and resets the PSD
-        
-        Parameters
-        ----------
-        min : float
-            Lower bound of PSD
-        cMax : float
-            Upper bound of PSD
-        bins : int
-            Number of bins
-        linearSpacing : bool (optional)
-            If True (default), will store PSD on a linear scale
-            Else, will store PSD on a logarithmic scale
-        resetPSD : bool (optional)
-            Whether to reset the PSD (defaults to True)
+        For Euler implementation, sets whether to change the bin size when 
+        the number of filled bins > maxBins or < minBins
+
+        If False, the bins will still change if nucleated particles are greater than the max bin size
+        and bins will still be added when the last bins starts to fill (although this will not change the bin size)
         '''
+        self._adaptiveBinSize = adaptive
+
+    def setBinConstraints(self, bins = 150, minBins = 100, maxBins = 200):
+        '''
+        Sets constraints for minimum and maxinum number of bins over an order of magnitude
+
+        All bins will be overridden to an even number
+        Minimum number of bins will be overridden to be at most half of maximum bins
+        Default bins will be overridden to be halfway between min and max bins if out of range
+        '''
+        self.minBins = minBins
+        self.maxBins = maxBins
         self.bins = bins
-        self.min = cMin
-        self.max = cMax
-        self.linear = linearSpacing
-        
-        if linearSpacing:
-            self.PSDbounds = np.linspace(cMin, cMax, bins+1)
-            self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
-        else:
-            self.PSDbounds = np.logspace(cMin, cMax, bins+1)
-            self.PSDsize = np.power(10, 0.5 * (np.log(self.PSDbounds[1:] * self.PSDbounds[:-1])))
-            
-        if resetPSD:
-            self.reset()
 
-    def addSizeClass(self):
-        '''
-        Adds an additional size class to end of distribution
-        '''
-        self.bins += 1
-        self.PSDbounds = np.append(self.PSDbounds, 0)
-        self.PSDsize = np.append(self.PSDsize, 0)
-        self.PSD = np.append(self.PSD, 0)
-
-        if self.linear:
-            self.max += (self.PSDbounds[1] - self.PSDbounds[0])
-            self.PSDbounds[-1] = self.max
-            self.PSDsize[-1] = 0.5 * (self.PSDbounds[-1] + self.PSDbounds[-2])
-        else:
-            self.max = np.power(10, np.log(self.max * self.PSDbounds[1] / self.PSDbounds[0]))
-            self.PSDbounds[-1] = self.max
-            self.PSDsize[-1] = np.power(10, 0.5 * np.log(self.PSDbounds[-1] * self.PSDbounds[-2]))
-        
     def LoadDistribution(self, data):
         '''
         Creates a particle size distribution from a set of data
@@ -118,13 +106,157 @@ class PopulationBalanceModel:
         '''
         self.PSD, self.PSDbounds = np.histogram(data, self.PSDbounds)
         self.PSD = self.PSD.astype('float')
+
+    def LoadDistributionFunction(self, function):
+        '''
+        Creates a particle size distribution from a function
+
+        Parameters
+        ----------
+        function : function
+            Takes in R and returns density
+        '''
+        self.PSD = function(self.PSDsize)
+
+    def createBackup(self):
+        '''
+        Stores current PSD and PSDbounds
+        '''
+        self._prevPSD = copy.copy(self.PSD)
+        self._prevPSDbounds = copy.copy(self.PSDbounds)
+
+    def revert(self):
+        '''
+        Reverts to previous PSD and PSDbounds
+        '''
+        self.PSD = copy.copy(self._prevPSD)
+        self.PSDbounds = copy.copy(self._prevPSDbounds)
+        self.PSDsize = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
+        self.bins = len(self.PSD)
+        self.min, self.max = self.PSDbounds[0], self.PSDbounds[-1]
+
+    def reset(self, resetBounds = True):
+        '''
+        Resets the PSD to 0
+        This will remove any size classes that were added since initialization
+        '''
+        if resetBounds:
+            self.min = self.originalMin
+            self.max = self.originalMax
+            self.bins = self.originalBins
+        self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
+        self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
+            
+        self.PSD = np.zeros(self.bins)
+
+    def changeSizeClasses(self, cMin, cMax, bins = None, resetPSD = False):
+        '''
+        Changes the size classes and resets the PSD
+        
+        Parameters
+        ----------
+        cMin : float
+            Lower bound of PSD
+        cMax : float
+            Upper bound of PSD
+        bins : int
+            Number of bins
+        resetPSD : bool (optional)
+            Whether to reset the PSD (defaults to False)
+        '''
+        self.bins = self.bins if bins is None else bins
+        self.min = cMin
+        self.max = np.amax([10*self.min, cMax])
+
+        if resetPSD:
+            self.reset()
+        else:
+            oldV = self.ThirdMoment()
+            distDen = self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])
+            rOld = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
+            self.reset(False)
+            self.PSD = np.interp(self.PSDsize, rOld, distDen) * (self.PSDbounds[1:] - self.PSDbounds[:-1])
+            newV = self.ThirdMoment()
+            if newV != 0:
+                self.PSD *= oldV / newV
+            else:
+                self.PSD = np.zeros(self.bins)
+
+    def addSizeClasses(self, bins = 1):
+        '''
+        Adds an additional size class to end of distribution
+
+        Parameters
+        ----------
+        bins : int
+            Number of bins to add
+        '''
+        self.bins += bins
+        self.PSD = np.append(self.PSD, np.zeros(bins))
+
+        self.max += bins * (self.PSDbounds[1] - self.PSDbounds[0])
+        self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
+        self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
+
+    def getDTEuler(self, currDT, growth, maxDissolution, startIndex):
+        '''
+        Calculates time interval for Euler implementation
+            dt < dR / (2 * growth rate)
+        This ensures that at most, only half of particles in one size class can go to another
+
+        Parameters
+        ----------
+        currDT : float
+            Current time interval, will be returned if it's smaller than what's given by the contraint
+        growth : array of floats
+            Growth rate, must have lenth of bins+1 (or length of PSDbounds)
+        maxDissolution : float
+            Maximum volume allowed to dissolve
+        startIndex : int
+            First index to look at for growth rate, all indices below startIndex will be ignored
+        '''
+        dissFrac = maxDissolution * self.ThirdMoment()
+        dissIndex = np.amax([np.argmax(self.CumulativeMoment(3) > dissFrac), startIndex])
+        growthFilter = growth[dissIndex:-1][self.PSD[dissIndex:] > 0]
+        #if len(growthFilter) == 0 or np.amax(growthFilter) < 0:
+        if len(growthFilter) == 0:
+            return currDT
+        else:
+            return np.amin([currDT, (self.PSDbounds[1] - self.PSDbounds[0]) / (2 * np.amax(np.abs(growthFilter)))])
+
+    def adjustSizeClassesEuler(self, checkDissolution = False):
+        '''
+        Adds a size class if last class in PBM is filled
+        Changes length of size classes based off number of allowed bins
+        '''
+        change = False
+        newIndices = None
+        if self.PSD[-1] > 1:
+            #print('adding bins')
+            newIndices = self.bins
+            self.addSizeClasses(int(self.originalBins/4))
+            change = True
+
+        if self._adaptiveBinSize:
+            if self.bins > self.maxBins:
+                #print('reducing bins')
+                self.changeSizeClasses(self.PSDbounds[0], self.PSDbounds[-1], self.minBins)
+                change = True
+                newIndices = None
+            elif checkDissolution and self.PSDbounds[-1] > 10*self.PSDbounds[0]:
+                if any(self.PSD > 1) and np.amax(self.PSDsize[self.PSD > 1]) < self.PSDsize[int(self.minBins/2)]:
+                    #print('splitting bins')
+                    self.changeSizeClasses(self.PSDbounds[0], np.amax(self.PSDbounds[1:][self.PSD > 1]), self.maxBins)
+                    change = True
+                    newIndices = None
+        return change, newIndices
         
     def Normalize(self):
         '''
         Normalizes the PSD to 1
         '''
-        sum = self.WeightedMoment(0, self.PSDbounds[1:] - self.PSDbounds[:-1])
-        self.PSD /= sum
+        total = self.WeightedMoment(0, self.PSDbounds[1:] - self.PSDbounds[:-1])
+        self.PSD /= total
             
     def NormalizeToMoment(self, order = 0):
         '''
@@ -136,10 +268,38 @@ class PopulationBalanceModel:
             Order of moment that PSD will be normalized to (defaults to 0)
             Using zeroth order will normalize the PSD so the sum of PSD will be 1
         '''
-        sum = self.Moment(order)
-        self.PSD /= sum
+        total = self.Moment(order)
+        self.PSD /= total
+
+    def Nucleate(self, amount, radius):
+        '''
+        Adds nucleated particles to PSD given radius and amount of particles
+
+        Parameters
+        ----------
+        amount : float
+            Amount of nucleated particles
+        radius : float
+            Radius of nucleated particles
+        '''
+        if amount < 1:
+            return False
+            
+        change = False
+
+        #Find size class for nucleated particles
+        nRad = np.argmax(self.PSDbounds > radius) - 1
+
+        #If radius is larger than length scale of PBM, adjust PBM such that radius is towards the beginning
+        if nRad == -1 and radius > 0:
+            #print('adding nucleated bins')
+            self.changeSizeClasses(self.PSDbounds[0], 5 * radius, self.originalBins)
+            nRad = np.argmax(self.PSDbounds > radius)
+            change = True
+        self.PSD[nRad] += amount
+        return change
         
-    def Update(self, dt, flux, externalRates = None):
+    def UpdateEuler(self, dt, flux):
         '''
         Updates PSD given the flux and any external contributions
 
@@ -153,45 +313,42 @@ class PopulationBalanceModel:
         flux : array
             Growth rate of each particle size class
             Array size must be (bins + 1) since this operates on bounds of size classes
-        externalRates : array (optional)
-            Additional contribution of particles (nucleation or dissolution)
-            Array size must be (bins) since this operates on the individual size classes
         '''
-        #Store current PSD
-        self._prevPSD = copy.copy(self.PSD)
-
         netFlux = np.zeros(self.bins + 1)
-        netFlux[0] = 0 if flux[0] > 0 else flux[0] * dt * self.PSD[0] / (self.PSDbounds[1] - self.PSDbounds[0])
-        netFlux[-1] = 0 if flux[-1] < 0 else flux[-1] * dt * self.PSD[-1] / (self.PSDbounds[-1] - self.PSDbounds[-2])
 
-        #If flux is going from size class n to n-1, then use size class n (flux <= 0)
-        indices = flux[1:-1] <= 0
-        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[1:][indices] / (self.PSDbounds[2:] - self.PSDbounds[1:-1])[indices]
-        under = (netFlux[1:-1] < -self.PSD[1:]) & indices
-        netFlux[1:-1][under] = -self.PSD[1:][under]
-        
-        #if flux is going from size class n-1 to n, then use size class n-1 (flux > 0)
-        indices = ~indices
-        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[:-1][indices] / (self.PSDbounds[1:-1] - self.PSDbounds[:-2])[indices]
-        over = (netFlux[1:-1] > self.PSD[:-1]) & indices
-        netFlux[1:-1][over] = self.PSD[:-1][over]
+        #Array of 0 (flux <= 0), 1 (flux > 0)
+        fluxSign = np.sign(flux)
+        fluxSign[fluxSign == -1] = 0
+
+        #If flux is negative (from class n to n-1), then take from size class n
+        #If flux is positive, then take from size class n-1
+        netFlux[:-1] += dt * flux[:-1] * self.PSD * (1-fluxSign[:-1]) / (self.PSDbounds[1:] - self.PSDbounds[:-1])
+        netFlux[1:] += dt * flux[1:] * self.PSD * fluxSign[1:] / (self.PSDbounds[1:] - self.PSDbounds[:-1])
+
+        #Brute force stability so PSD is never negative
+        #If the time step is determined from getDTEuler, this should be unnecessary
+        netFlux[1:-1][netFlux[1:-1] < -self.PSD[1:]] = -self.PSD[1:][netFlux[1:-1] < -self.PSD[1:]]
+        netFlux[1:-1][netFlux[1:-1] > self.PSD[:-1]] = self.PSD[:-1][netFlux[1:-1] > self.PSD[:-1]]
 
         self._fv = netFlux
         
-        self.PSD += netFlux[:-1] - netFlux[1:]
-        if externalRates is not None:
-            self.PSD += externalRates
+        self.PSD += (netFlux[:-1] - netFlux[1:])
+        
+        #Adjust size classes and return True if the size classes had changed
+        change, newIndices = self.adjustSizeClassesEuler(all(flux<0))
             
         #Set negative frequencies to 0
         self.PSD[self.PSD < 1] = 0
 
-    def UpdateMomentConserved(self, dt, flux, externalRates = None, order = 1):
+        return change, newIndices
+
+    def UpdateLagrange(self, dt, flux):
         '''
-        Updates PSD given the flux and any external contributions
+        Updates bounds of size classes with given growth rate
+        Fluxes of particles between size classes is d(Gn)/dx,
+        however, keeping the number of particles in each size class the same,
+        the bounds of the size classes can be updated by r_i = v_i * dt
 
-        Change in the amount of particles in a given size class = d(G*n^m)/dx
-        Where G is flux of size class, n is number of particles in size class, m is moment order and dx is range of size class
-        
         Parameters
         ----------
         dt : float
@@ -199,40 +356,10 @@ class PopulationBalanceModel:
         flux : array
             Growth rate of each particle size class
             Array size must be (bins + 1) since this operates on bounds of size classes
-        externalRates : array (optional)
-            Additional contribution of particles (nucleation or dissolution)
-            Array size must be (bins) since this operates on the individual size classes
-        order : int (optional)
-            Order of moment to conserve (defaults to 1, which is the same as the Update method)
         '''
-        #Store current PSD
-        self._prevPSD = copy.copy(self.PSD)
-        
-        netFlux = np.zeros(self.bins + 1)
-        netFlux[0] = 0 if flux[0] > 0 else flux[0] * dt * self.PSD[0] * self.PSDsize[0]**order / (self.PSDbounds[1] - self.PSDbounds[0])
-        netFlux[-1] = 0 if flux[-1] < 0 else flux[-1] * dt * self.PSD[-1] * self.PSDsize[-1]**order / (self.PSDbounds[-1] - self.PSDbounds[-2])
-
-        #Note - this is not tested yet
-        #If flux is going from size class n to n-1, then use size class n (flux <= 0)
-        indices = flux[1:-1] <= 0
-        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[1:][indices] * self.PSDsize[1:][indices]**order / (self.PSDbounds[2:] - self.PSDbounds[1:-1])[indices]
-        under = (netFlux[1:-1] < -self.PSD[1:] * self.PSDsize[1:]**order) & indices
-        netFlux[1:-1][under] = -self.PSD[1:][under] * self.PSDsize[1:][under]**order
-        
-        #if flux is going from size class n-1 to n, then use size class n-1 (flux > 0)
-        indices = ~indices
-        netFlux[1:-1][indices] = dt * flux[1:-1][indices] * self.PSD[:-1][indices] * self.PSDsize[:-1][indices]**order / (self.PSDbounds[1:-1] - self.PSDbounds[:-2])[indices]
-        over = (netFlux[1:-1] > self.PSD[:-1] * self.PSDsize[:-1]**order) & indices
-        netFlux[1:-1][over] = self.PSD[:-1][over] * self.PSDsize[1:][over]**order
-
-        self._fv = netFlux
-        
-        self.PSD += (netFlux[:-1] - netFlux[1:]) / self.PSDsize**order
-        if externalRates is not None:
-            self.PSD += externalRates
-            
-        #Set negative frequencies to 0
-        self.PSD[self.PSD < 0] = 0
+        self._prevPSDbounds = copy.copy(self.PSDbounds)
+        self.PSDbounds += flux * dt
+        self.PSDsize = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
         
     def Moment(self, order):
         '''
@@ -288,25 +415,25 @@ class PopulationBalanceModel:
         '''
         Sum of the PSD
         '''
-        return np.sum(self.PSD)
+        return self.Moment(0)
     
     def FirstMoment(self):
         '''
         Length weighted moment
         '''
-        return np.sum(self.PSD * self.PSDsize)
+        return self.Moment(1)
         
     def SecondMoment(self):
         '''
         Area weighted moment
         '''
-        return np.sum(self.PSD * self.PSDsize**2)
+        return self.Moment(2)
         
     def ThirdMoment(self):
         '''
         Volume weighted moment
         '''
-        return np.sum(self.PSD * self.PSDsize**3)
+        return self.Moment(3)
         
     def PlotCurve(self, axes, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
         '''
@@ -329,6 +456,9 @@ class PopulationBalanceModel:
                 determined by precipitate curvature
         *args, **kwargs - extra arguments for plotting
         '''
+        if hasattr(scale, '__len__'):
+            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
+
         if fill:
             axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
         else:
@@ -357,6 +487,9 @@ class PopulationBalanceModel:
                 determined by precipitate curvature
         *args, **kwargs - extra arguments for plotting
         '''
+        if hasattr(scale, '__len__'):
+            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
+
         if fill:
             axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
         else:
@@ -374,10 +507,10 @@ class PopulationBalanceModel:
 
         #Set y-limits
         if logY:
-            axes.set_ylim([1e-1, np.amax([1.1 * np.max(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
+            axes.set_ylim([1e-1, np.amax([1.1 * np.amax(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
             axes.set_yscale('log')
         else:
-            axes.set_ylim([0, np.amax([1.1 * np.max(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
+            axes.set_ylim([0, np.amax([1.1 * np.amax(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
 
     def PlotKDE(self, axes, bw_method = None, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
         '''
@@ -404,8 +537,11 @@ class PopulationBalanceModel:
         *args, **kwargs - extra arguments for plotting
         '''
         kernel = sts.gaussian_kde(self.PSDsize, bw_method = bw_method, weights = self.PSD)
-        x = np.linspace(self.min, self.max, 1000) if self.linear else np.logspace(self.min, self.max, 1000)    
+        x = np.linspace(self.min, self.max, 1000)   
         y = kernel(x) * self.ZeroMoment() * (self.PSDbounds[1] - self.PSDbounds[0])
+
+        if hasattr(scale, '__len__'):
+            scale = np.interp(x, self.PSDbounds, scale)
         
         if fill:
             axes.fill_between(x * scale, y, np.zeros(len(y)), *args, **kwargs)
@@ -446,6 +582,9 @@ class PopulationBalanceModel:
             xCoord[0], xCoord[1::2], xCoord[2::2] = self.PSDbounds[0], self.PSDbounds[:-1], self.PSDbounds[1:]
             yCoord[1::2], yCoord[2::2] = self.PSD, self.PSD
 
+        if hasattr(scale, '__len__'):
+            scale = np.interp(xCoord, self.PSDbounds, scale)
+
         if outline != 'no outline':
             axes.plot(xCoord * scale, yCoord, *args, **kwargs)
             if fill:
@@ -473,6 +612,9 @@ class PopulationBalanceModel:
             Moment of specified order
         *args, **kwargs - extra arguments for plotting
         '''
+        if hasattr(scale, '__len__'):
+            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
+
         axes.plot(self.PSDsize * scale, self.CumulativeMoment(order) / self.Moment(order), *args, **kwargs)
         self.setAxes(axes, logX, False) 
         axes.set_ylim([0, 1])
@@ -502,10 +644,10 @@ class PopulationBalanceModel:
         #Don't set y limits if the PSD is empty
         if any(self.PSD > 0): 
             if logY:
-                axes.set_ylim([1e-1, np.amax([1.1 * np.max(self.PSD), 1])])
+                axes.set_ylim([1e-1, np.amax([1.1 * np.amax(self.PSD), 1])])
                 axes.set_yscale('log')
             else:
-                axes.set_ylim([0, np.amax([1.1 * np.max(self.PSD), 1])])
+                axes.set_ylim([0, np.amax([1.1 * np.amax(self.PSD), 1])])
         
         axes.set_ylabel('Frequency')
             
